/* *****************************************************************************
 *
 * Filename: CalendarBase.java
 *
 * Author:   Robert Thornburrow
 * Project:  CalendarMaker
 * Created:  17 December 2002, 11:33:13
 *
 * Purpose:  Base class for all Calendar beans (Table, List, etc.).
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

package com.tsac.calmaker.beans;

import javax.swing.*;
import javax.accessibility.*;
import java.awt.*;
import java.awt.event.*;

import java.util.*;
import java.io.*;
import java.text.*;

import com.tsac.calmaker.*;

public abstract class CalendarBase
    extends javax.swing.JComponent
    implements IDateable, IScaleable, Accessible
{
    // Statics
    protected final ResourceBundle resources = ResourceBundle.getBundle( "com/tsac/calmaker/beans/CalendarBeans" );

    // properties
    protected Color _cellbordercolor;
    protected Color _celltextcolor;
    protected Color _cellwendtextcolor;
    protected Font _dayfont;
    protected Font _datefont;
    protected Font _diaryfont;
    protected Calendar _dispdate = new GregorianCalendar();
    protected int _inset;
    protected int _botmargin;
            
    // All holdiay entries keyed by year
    protected Hashtable _holidays = new Hashtable();
    
    // All diary entries keyed by year
    protected Hashtable _dairycache = new Hashtable();
    
    // runtime state
    protected CalendarCell[] _diary = null;   // diary entries for the current month
    protected int[] _cells = new int[31];     // diary entry position info
    protected int _dispyear = -1;             // current year (so holidays get loaded)
    protected Font _actualdayfont;
    protected Font _actualdatefont;
    protected Font _actualdiaryfont;
    protected int _cellwidth;        // calculated
    protected int _cellheight;       // calculated
    protected int _leftmargin;       // calculated
    protected FontMetrics _dayfontmetrics;
    protected FontMetrics _datefontmetrics;
    protected FontMetrics _diaryfontmetrics;
    protected boolean _useshortdaynames;
    protected int _firstday;
    protected int _currmonth;
    protected boolean _resized = true;


    ///////////////////////////////////////////////////////////////////////////
    // public interface

    // Constructor
    public CalendarBase()
    {
        // defaults for the color properties
        _cellbordercolor = Color.black;
        _celltextcolor   = Color.black; //darkGray;
        _cellwendtextcolor = Color.red;
        
        // defaults for font props (defaults to 12pt Helvetica)
        _dayfont   = new Font( "Helvetica", Font.PLAIN, 14 );
        _datefont  = _dayfont;
        _diaryfont = new Font( "Helvetica", Font.PLAIN, 7 );

        // other props
        _inset = 0;
        _botmargin = 0;
        
        setDate( new Date() );
    }
    
    
    public void setBounds( int x, int y, int w, int h )
    {
        _resized = true;
        super.setBounds( x, y, w, h );
        calclayout();
        repaint();
    }
    
    
    // Returns the current date being displayed by the bean
    public Date getDate()
    {
        return _dispdate.getTime();
    }

    // Sets the current date being displayed by the bean
    public void setDate( int year, int month )
    {
        _dispdate.set( year, month, 1 );
        setDate( _dispdate );
    }

    // Sets the current date being displayed by the bean
    public void setDate( Date date )
    {
        _dispdate.setTime( date );
        setDate( _dispdate );
    }
    
    // Sets the current date being displayed by the bean
    private void setDate( Calendar cal )
    {
        _dispdate = cal;
        _dispdate.set( Calendar.DAY_OF_MONTH, 1 );
        
        int dispyear = _dispdate.get( Calendar.YEAR );
        if( dispyear != _dispyear )
        {
            loadHolidays( dispyear );
        }
        
        _dispyear = dispyear;
        _currmonth = _dispdate.get( Calendar.MONTH );
        _firstday = _dispdate.get( Calendar.DAY_OF_WEEK ) - 1;
        
        calcDiary();
        repaint();
    }

    
    ///////////////////////////////////////////////////////////////////////////
    // public properties
        
    public Color getCellBorderColor() { return _cellbordercolor; }
    public void setCellBorderColor( Color c ) { _cellbordercolor = c; repaint(); }

    public Color getCellTextColor() { return _celltextcolor; }
    public void setCellTextColor( Color c ) { _celltextcolor = c; repaint(); }

    public Color getCellWeekendTextColor() { return _cellwendtextcolor; }
    public void setCellWeekendTextColor( Color c ) { _cellwendtextcolor = c; repaint(); }

    public Font getDayFont() { return _dayfont; }
    public void setDayFont( Font f ) { _dayfont = f; _resized = true; repaint(); }

    public Font getDateFont() { return _datefont; }
    public void setDateFont( Font f ) { _datefont = f; _resized = true; repaint(); }

    public Font getDiaryFont() { return _diaryfont; }
    public void setDiaryFont( Font f ) { _diaryfont = f; _resized = true; repaint(); }

    public int getInset() { return _inset; }
    public void setInset( int in ) { _inset = in; _resized = true; repaint(); }

    public int getBottomMargin() { return _botmargin; }
    public void setBottomMargin( int m ) { _botmargin = m; _resized = true; repaint(); }


    ///////////////////////////////////////////////////////////////////////////
    // private implementation

    protected abstract void calclayout();

    
    public abstract void paint( Graphics g, double scale );
    
    
    public void paint( Graphics g )
    {
        paint( g, 1.0 );
    }
    
    
    protected int getDaysInMonth()
    {
        int i = _dispdate.get( Calendar.MONTH );
        int j = _dispdate.get( Calendar.YEAR );
        int an[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
        int k = an[i];
        if (i == 1 && j % 4 == 0)
            k++;
        return k;
    }


    protected String getMonthName()
    {
        int i = _dispdate.get( Calendar.MONTH );
        return Constants.months[i];
    }

    private String getYear()
    {
        int i = _dispdate.get( Calendar.YEAR );
        return Integer.toString( i );
    }
    
    public void calcDiary()
    {
        _diary = null;
        
        for( int n = 0; n < 31; n++ )
            _cells[n] = 0;

        // put in holidays first
        CalendarCell[][] holsyear = (CalendarCell[][]) _holidays.get( new Integer( _dispdate.get( Calendar.YEAR ) ) );

        if( holsyear != null )
        {
            CalendarCell[] hols = holsyear[ _dispdate.get( Calendar.MONTH ) ];
            
            if( hols != null )
            {
                for( int n = 0; n < hols.length; n++ )
                    addDiaryEntry( hols[n] );
            }
        }
    }
    
    protected void addDiaryEntry( CalendarCell entry )
    {
        if( _diary == null )
            _diary = new CalendarCell[this.getDaysInMonth()];
        
        int startdate = entry.getDate();
        
        int start = startdate-1;
        
        _cells[start]++;
        _diary[start] = entry;
        entry.setPos( _cells[start] );
    }

    public void update( Graphics g )
    {
        paint(g);
    }
    
    
    protected String clipString( String str, int cur, int des )
    {
        int clip = (str.length() * des / cur) - 2;
        return str.substring( 0, clip ) + "...";
    }

    
    ///////////////////////////////////////////////////////////////////////////
    
    protected void loadHolidays( int year )
    {
        String filename = Integer.toString( year ) + "holidays.csv";
        
        try
        {
            FileReader file = new FileReader( filename );
            BufferedReader in = new BufferedReader( file );
            
            loadHolidays( year, in );
        }
        catch( FileNotFoundException e )
        {
            System.out.println( "Missing holdiay file: " + filename );
        }
        catch( IOException e )
        {
            e.printStackTrace();
        }
    }
    
    public void loadHolidays( int year, BufferedReader in )
    {
        try
        {
            CalendarCell[][] hols = new CalendarCell[12][];
            Vector entries = new Vector();
            String line;
            int month = -1;
            
            while( (line = in.readLine()) != null )
            {
                //System.out.println( line );
                
                if( (line.length() == 0) || (line.charAt( 0 ) == '#') )
                    continue;
                    
                if( line.charAt( 0 ) == '@' )
                {
                    if( (entries.size() > 0) && (month >= 0) && (month < 12) )
                    {
                        hols[month] = new CalendarCell[ entries.size() ];
                        entries.copyInto( hols[month] );
                        entries.removeAllElements();
                    }
                    month++;
                    continue;
                }
                
                StringTokenizer tok = new StringTokenizer( line, "," );
                
                try
                {
                    String datestr = tok.nextToken();
                    String what = tok.nextToken();
                    
                    int date = Integer.parseInt( datestr );
                    
                    entries.addElement( new CalendarCell( date, what ) );
                }
                catch( NumberFormatException e )
                {
                    System.out.println( "Corrupt holiday file (date), line: " + line );
                }
                catch( NoSuchElementException e )
                {
                    System.out.println( "Corrupt holiday file (format), line: " + line );
                }
            }// while

            if( (entries.size() > 0) && (month >= 0) && (month < 12) )
            {
                hols[month] = new CalendarCell[ entries.size() ];
                entries.copyInto( hols[month] );
            }

            System.out.println( "Holidays for year " + year + " loaded." );

            _holidays.put( new Integer( year ), hols );
        }
        catch( IOException e )
        {
            e.printStackTrace();
        }
    }
        
    protected void checkComma( StringTokenizer tok, int n )
    {
        if( !tok.nextToken().equals( "," ) )
            throw new NoSuchElementException( "Missing , " + n );
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Accessibility stuff
    
    protected abstract String getAccessibleName();
    
    protected abstract String getAccessibleDescription();
    
    
    public class AccessibleCalendarBase
        extends JComponent.AccessibleJComponent
    {
        public String getAccessibleName()
        {
            // Use name set by the developer if set, otherwise use name from bean
            String name = super.getAccessibleName();            
            return (name != null) ? name : CalendarBase.this.getAccessibleName();
        }
        
        public String getAccessibleDescription()
        {
            // Use description set by the developer if set, otherwise use description from bean
            String name = super.getAccessibleDescription();            
            return (name != null) ? name : CalendarBase.this.getAccessibleDescription();
        }
    }
    
    
    public AccessibleContext getAccessibleContext()
    {
        if( accessibleContext == null )
            accessibleContext = new AccessibleCalendarBase();
        
        return accessibleContext;
    }
}