/* *****************************************************************************
 *
 * Filename: DateablePanel.java
 *
 * Author:   Robert Thornburrow
 * Project:  CalendarMaker
 * Created:  17 December 2002, 12:19
 *
 * Purpose:  Container for all CalendarMaker beans.  Does much of the work of
 * selecting/resizing/moving and inserting beans.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

package com.tsac.calmaker.beans;

import javax.swing.*;
import javax.swing.event.EventListenerList;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.awt.print.*;
import java.io.File;
import java.util.*;

import com.tsac.calmaker.IDateable;
import com.tsac.calmaker.IScaleable;
import com.tsac.calmaker.CalendarPanel;

public class DateablePanel
    extends javax.swing.JPanel
    implements IDateable, Printable, Autoscroll,
               MouseListener, MouseMotionListener
{
    // Constants
    public static final int margin = 12;
    
    // Properties
    protected int _width = 768;
    protected int _height = 1024;
    protected int _numPagesWidth = 1;
    protected EventListenerList _listenerList = new EventListenerList();
    
    // State
    protected transient int _month;
    protected transient int _year;
    protected transient Component _selection = null;
    protected transient Class _selBean = null;
    protected transient double _scale = 1.0;
    protected transient double _iscale = 1.0;
    
    // Cached stuff
    protected transient ResourceBundle calendarBeans = ResourceBundle.getBundle( "com/tsac/calmaker/beans/CalendarBeans" );
    protected transient int _cursor;
    protected transient boolean _moveable;
    protected transient Rectangle _bounds    = null;
    protected transient Rectangle _topleft   = null;
    protected transient Rectangle _topright  = null;
    protected transient Rectangle _botleft   = null;
    protected transient Rectangle _botright  = null;
    protected transient Rectangle _topcenter = null;
    protected transient Rectangle _midleft   = null;
    protected transient Rectangle _botcenter = null;
    protected transient Rectangle _midright  = null;
    protected transient Point _dragStart = null;
    protected transient int _dragMode;
    protected transient double _dragRatio;
    protected transient DPDropTarget _dropTarget;
    protected transient Component _dropSite;
    
    
	public DateablePanel()
	{
        // Deal with selecting components
        this.addMouseListener( this );
        
        // Deal with changing the cursor and drags
        this.addMouseMotionListener( this );
        
        // Deal with this component being moved or resized
        this.addComponentListener( rectangleUpdater );
        
        // Deal with new components being added
        this.addContainerListener( new ComponentUpdater() );
        
        // Make this a focus cycle so we can use tab to select components
        this.addFocusListener( selectionUpdater );
        setFocusCycleRoot( true );
        setFocusable( true );
        
        // Attach handler for DnD drop events
        _dropTarget = new DPDropTarget( this );
    }
    
    
    public void setNumPagesWidth( int p )  { _numPagesWidth = p; }
    public int getNumPagesWidth()  { return _numPagesWidth; }

    
    public Component getSelection()  { return _selection; }
    
    public void setSelection( Component c )
    {
        if( _selection != c )
        {
            _selection = c;
            updateSelectionRectangles();
            fireDateableSelectionEvent( new DateableSelectionEvent( DateablePanel.this, _selection, DateableSelectionEvent.SELECTION_CHANGED ) );
            repaint();
        }
    }

    
    public void setZoom( double scale )
    {
        _scale = scale;
        _iscale = (_scale != 0 ? 1/_scale : 1.0);        
        updateSelectionRectangles();
    }
    

    // Set the actual size of the panel
    public void setActualSize( Dimension sz )
    {
        this.setActualSize( sz.width, sz.height );
    }
    

    // Set the actual size of the panel
    public void setActualSize( int w, int h )
    {
        _width = w;
        _height = h;
        
        this.setSize( w, h );
    }
    
    
    // Return the actual size
    public Dimension getActualSize()
    {
        return new Dimension( _width, _height );
    }

    
    // Override default
    public Dimension getPreferredSize()
    {
        return new Dimension( _width, _height );
    }
    
    
    // Override default
    public Dimension getMaximumSize()
    {
        return new Dimension( _width, _height );
    }
    
    
    // Override default
    public void removeAll()
    {
        super.removeAll();
        setSelection( null );
    }
    
    
    // IDateable
    public void setDate( int year, int month )
    {
        _year = year;
        _month = month;
        
		for( int n = 0; n < getComponentCount(); n++ )
		{
		    Component c = getComponent( n );
		    
		    if( c instanceof IDateable )
		        ((IDateable) c).setDate( year, month );
		}
    }
    
    
    public void setSelectedBean( Class beanClass )
    {
        _selBean = beanClass;
        
        if( beanClass != null )
            setSelection( null );
    }
    
    
    public void addDateableSelectionListener( DateableSelectionListener l )
    {
        _listenerList.add( DateableSelectionListener.class, l );
    }

    
    public void removeDateableSelectionListener( DateableSelectionListener l )
    {
        _listenerList.remove( DateableSelectionListener.class, l );
    }
    
    
    /**
     * Fires any of DateableSelectionEvent to all the listeners.
     */
    protected void fireDateableSelectionEvent( DateableSelectionEvent e )
    {
        Object[] listeners = _listenerList.getListenerList();

        for( int i = listeners.length-2; i >= 0; i -= 2 )
        {
            if( listeners[i] == DateableSelectionListener.class )
            {
                switch( e.getKind() )
                {
                case DateableSelectionEvent.SELECTION_CHANGED:
                    ((DateableSelectionListener) listeners[i+1]).selectionChanged( e );
                    break;
                case DateableSelectionEvent.IMPORT_TEMPLATE:
                    ((DateableSelectionListener) listeners[i+1]).importTemplate( e );
                    break;
                case DateableSelectionEvent.DOUBLE_CLICKED:
                    ((DateableSelectionListener) listeners[i+1]).doubleClicked( e );
                    break;
                }
            }
        }
    }
    
    
    // Paint hook for drawing selection rectangles
    public void paint( Graphics g )
    {
        // Draw all the components underneath before overlaying selection etc.
        super.paint( g );
        
        // Draw the autoscroll margins
        //g.setColor( Color.red );
        //Insets asin = this.getAutoscrollInsets();
        //g.drawRect( asin.left, asin.top, getWidth() - asin.left - asin.right, getHeight() - asin.top - asin.bottom );
        
        // Draw the drop site rectangle
        if( _dropSite != null )
        {
            Rectangle bounds = SwingUtilities.convertRectangle( _dropSite.getParent(), _dropSite.getBounds(), this );
            int x = (int) (_scale * bounds.x);
            int y = (int) (_scale * bounds.y);
            int width  = (int) (_scale * bounds.width);
            int height = (int) (_scale * bounds.height);
            
            g.setColor( Color.blue );
            g.drawRect( x, y, width, height );      // bounding box
            
            return;     // Omit the selection rectangle
        }
        
        // Draw the selection rectangle
        if( _selection != null )
        {
            Rectangle bounds = SwingUtilities.convertRectangle( _selection.getParent(), _selection.getBounds(), this );
            int x = (int) (_scale * bounds.x);
            int y = (int) (_scale * bounds.y);
            int width  = (int) (_scale * bounds.width);
            int height = (int) (_scale * bounds.height);
            int maxx = (int) (_scale * (bounds.x + bounds.width));
            int maxy = (int) (_scale * (bounds.y + bounds.height));
            int cenx = (int) (_scale * (bounds.x + bounds.width/2));
            int ceny = (int) (_scale * (bounds.y + bounds.height/2));
            
            g.setColor( _moveable ? Color.black : Color.lightGray );
            g.drawRect( x, y, width - 1, height - 1 );      // bounding box
            
            // corner handles
            g.fillRect( x, y, 5, 5 );               // top-left
            g.fillRect( maxx - 5, y, 5, 5 );        // top-right
            g.fillRect( x, maxy - 5, 5, 5 );        // bottom-left
            g.fillRect( maxx - 5, maxy - 5, 5, 5 ); // bottom-right
            
            // side handles
            g.fillRect( cenx - 2, y, 5, 5 );        // top-center
            g.fillRect( x, ceny - 2, 5, 5 );        // middle-left
            g.fillRect( cenx - 2, maxy - 5, 5, 5 ); // bottom-center
            g.fillRect( maxx - 5, ceny - 2, 5, 5 ); // middle-right
        }
    }
    
    
    // Scaled page paint
    public void paint( Graphics g, double scale )
    {
		for( int n = 0; n < getComponentCount(); n++ )
		{
		    Component c = getComponent( n );
    		        
		    if( c instanceof IScaleable )
		        ((IScaleable) c).paint( g, scale );
		}
    }
    
    public int print( Graphics jg, PageFormat pageFormat, int pi )
        throws PrinterException
    {
        if( pi >= 1 )
            return Printable.NO_SUCH_PAGE;
            
        double topMargin = pageFormat.getImageableY();
        double leftMargin = pageFormat.getImageableX();
        double bottomMargin = pageFormat.getHeight() - (pageFormat.getImageableY() + pageFormat.getImageableHeight());
        double rightMargin = pageFormat.getWidth() - (pageFormat.getImageableX() + pageFormat.getImageableWidth());
        
        System.out.println( "top = " + topMargin + ", left = " + leftMargin + ", right = " + rightMargin + ", bottom = " + bottomMargin );
        
        int pageWidth  = (int) (pageFormat.getWidth() / _numPagesWidth - (leftMargin + rightMargin));
        int pageHeight = (int) (pageFormat.getHeight() - (topMargin + bottomMargin));
        
        System.out.println( "pageWidth = " + pageWidth + "; pageHeight = " + pageHeight );
                       
        double hscale = (double) pageWidth / (double) _width;
        double vscale = (double) pageHeight / (double) _height;
                    
        double scale = Math.min( hscale, vscale );

        try
        {
            // One up page layout
            if( _numPagesWidth == 1 )
            {
                Graphics pg = jg.create( (int) leftMargin, (int) topMargin, pageWidth, pageHeight );
                              
                paint( pg, scale );
    		    
                pg.dispose();
            }
            else // two up page layout
            {
                Graphics pg1 = jg.create( (int) leftMargin, (int) topMargin, pageWidth, pageHeight );
                Graphics pg2 = jg.create( (int) (2*leftMargin + rightMargin) + pageWidth, (int) topMargin, pageWidth, pageHeight );
                
                int month = _month;
                
                // Draw first month
                setDate( _year, month );
                paint( pg1, scale );
                
                // Draw separator line
                jg.setColor( Color.lightGray );
                jg.drawLine(
                    (int) (leftMargin + rightMargin) + pageWidth, (int) topMargin,
                    (int) (leftMargin + rightMargin) + pageWidth, (int) topMargin + pageHeight );

                // Draw second month
                setDate( _year, month + 1 );
                paint( pg2, scale );
    		    
                // Reset date to current month
                setDate( _year, month );

                pg1.dispose();
                pg2.dispose();
            }
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }
        
        return Printable.PAGE_EXISTS;
    }
    
    
    protected void updateSelectionRectangles()
    {
        if( _selection == null )
            return;

        _moveable = !(_selection instanceof CalendarPanel) && (_selection != this);
        
        _bounds = SwingUtilities.convertRectangle( _selection.getParent(), _selection.getBounds(), this );

        _bounds.x = (int) (_scale * _bounds.x);
        _bounds.y = (int) (_scale * _bounds.y);
        _bounds.width  = (int) (_scale * _bounds.width);
        _bounds.height = (int) (_scale * _bounds.height);
        
        int maxx = _bounds.x + _bounds.width;
        int maxy = _bounds.y + _bounds.height;
        int cenx = _bounds.x + _bounds.width/2;
        int ceny = _bounds.y + _bounds.height/2;

        // corner handles
        _topleft  = new Rectangle( _bounds.x, _bounds.y, 5, 5 );
        _topright = new Rectangle( maxx - 5, _bounds.y, 5, 5 );
        _botleft  = new Rectangle( _bounds.x, maxy - 5, 5, 5 );
        _botright = new Rectangle( maxx - 5, maxy - 5, 5, 5 );

        // side handles
        _topcenter = new Rectangle( cenx - 2, _bounds.y, 5, 5 );
        _midleft   = new Rectangle( _bounds.x, ceny - 2, 5, 5 );
        _botcenter = new Rectangle( cenx - 2, maxy - 5, 5, 5 );
        _midright  = new Rectangle( maxx - 5, ceny - 2, 5, 5 );
    }
    
    
    /**
     * Work out where the mouse pointer is in relation to the selection
     * rectangle.  Returns one of the Cursor constants idicating that
     * the cursor is either inside/outside the rectangle, or over one of
     * the grab handles.
     */
    protected int cursorFromPosition( MouseEvent e, boolean move )
    {
        if( (_selection != null) && _moveable )
        {
            int x = e.getX();
            int y = e.getY();
            
            if( _topleft.contains( x, y ) )
                return Cursor.NW_RESIZE_CURSOR;
            else if( _topright.contains( x, y ) )
                return Cursor.NE_RESIZE_CURSOR;
            else if( _botleft.contains( x, y ) )
                return Cursor.SW_RESIZE_CURSOR;
            else if( _botright.contains( x, y ) )
                return Cursor.SE_RESIZE_CURSOR;
            else if( _topcenter.contains( x, y ) )
                return Cursor.N_RESIZE_CURSOR;
            else if( _midleft.contains( x, y ) )
                return Cursor.W_RESIZE_CURSOR;
            else if( _botcenter.contains( x, y ) )
                return Cursor.S_RESIZE_CURSOR;
            else if( _midright.contains( x, y ) )
                return Cursor.E_RESIZE_CURSOR;
            else if( move && _bounds.contains( x, y ) )
                return Cursor.MOVE_CURSOR;
        }
        
        return Cursor.DEFAULT_CURSOR;
    }
    
    
    /**
     * Create and insert a new instance of the given bean class, with the given
     * parent with its top-left corner at position x,y in the parents coordinate
     * space.
     */
    protected void insertBean( Class beanClass, Container parent, int x, int y )
    {
        try
        {
            // Create a new instance of beanClass
            Component bean = (Component) beanClass.newInstance();
            
            parent.add( bean );     // Add it to the parent
            
            // Set its initial size and location
            Dimension ps = bean.getPreferredSize();
            bean.setBounds( x, y, ps.width, ps.height );
            
            // Set its date if it is dateable
            if( bean instanceof IDateable )
                ((IDateable) bean).setDate( _year, _month );
            
            // Layout the parent (and any other containers)
            parent.validate();
            
            // Set it as the current selection
            setSelection( bean );
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }
    }
    
    
    /**
     * Setup variables we need to do a drag.  Need to record start point and
     * mode of drag, which will be used later when we actually do the resiziing
     * or moving.
     */
    protected void startDrag( Point start, int mode )
    {
        _dragStart = start;
        _dragMode = mode;
        
        if( _bounds != null )
            _dragRatio = (_bounds.height != 0) ? (_bounds.width / _bounds.height) : 1.0;
    }
    
    protected void endDrag()
    {
        _dragStart = null;
        _dragMode = Cursor.DEFAULT_CURSOR;
        
        if( _selection != null )
        {
            Container parent = _selection.getParent();
            
            if( parent != null )
                parent.validate();
        }
    }

    
    /**
     * Work out which component to use as a parent, based on the given point,
     * which is scaled using the current zoom setting.
     */
    private Container parentFromLocationForBeanInsert( Point loc )
    {
        // Get innermost component
        Component leaf = findComponentAt( (int) (_iscale * loc.x), (int) (_iscale * loc.y) );

        // Is it anything we want?
        if( leaf instanceof CalendarPanel )
            return (Container) leaf;

        // No, so try its parent
        Component parent = leaf.getParent();

        // Is parent acceptable?
        if( (parent != null) && (parent instanceof CalendarPanel) )
            return (Container) parent;

        return null;
    }

        
    /**
     * Handles mouse clicks, which either changes the current selection, or if a
     * bean has been selected in the toolbar, then that bean is inserted and the
     * toolbar bean selection is cleared.  In case of a double-click we simply
     * fire an event and return.
     */
    public void mouseClicked( MouseEvent e )
    {
        // If we were double clicked, generate an event
        if( e.getClickCount() == 2 )
        {
            fireDateableSelectionEvent( new DateableSelectionEvent( this, _selection, DateableSelectionEvent.DOUBLE_CLICKED ) );
            return;
        }
        
        // If we are in bean insert mode, try and do the insert
        if( _selBean != null )
        {
            Container parent = parentFromLocationForBeanInsert( e.getPoint() );
            if( parent != null )
            {
                insertBean( _selBean, parent, (int) (_iscale * e.getX()), (int) (_iscale * e.getY()) );
                setSelectedBean( null );
            }
            return;
        }
        
        // Determine component under mouse allowing for zoom setting
        Component newsel = findComponentAt( (int) (_iscale * e.getX()), (int) (_iscale * e.getY()) );
        
        // As long as it's not us, set the selection (DateablePanel is NOT selectable)
        if( newsel != this )
            setSelection( newsel );
        else
            setSelection( null );
    }
    
    
    public void mouseEntered( MouseEvent e )  {}
    
    public void mouseExited( MouseEvent e )  {}
    
    
    /**
     * Handles changing the mouse cursor when a mouse button is intially
     * pressed.  This is mainly to turn on the MOVE cursor if the cursor is
     * inside the selection rectangle.  Also sets the start point for a drag.
     */
    public void mousePressed( MouseEvent e )
    {
        int newcursor = cursorFromPosition( e, true );
        
        if( _cursor != newcursor )
        {
            setCursor( Cursor.getPredefinedCursor( newcursor ) );
            _cursor = newcursor;
        }
        
        startDrag( e.getPoint(), newcursor );
        
        requestFocus();
    }
    
    
    /**
     * Handles changing the mouse cursor when a mouse button is finally
     * released.  This is mainly to turn off the MOVE cursor if the cursor is
     * inside the selection rectangle.
     */
    public void mouseReleased( MouseEvent e )
    {
        int newcursor = cursorFromPosition( e, false );
        
        if( _cursor != newcursor )
        {
            setCursor( Cursor.getPredefinedCursor( newcursor ) );
            _cursor = newcursor;
        }
        
        endDrag();
    }
    
    
    private void checkFlipHoriz( Rectangle bounds, int newmode )
    {
        if( bounds.width < 0 )
        {
            bounds.width = - bounds.width;
            bounds.x -= bounds.width;
            _dragMode = newmode;
        }
    }

    private void checkFlipVert( Rectangle bounds, int newmode )
    {
        if( bounds.height < 0 )
        {
            bounds.height = - bounds.height;
            bounds.y -= bounds.height;
            _dragMode = newmode;
        }
    }
    
    private void checkFlipDiag( Rectangle bounds, int hmode, int vmode, int dmode )
    {
        if( (bounds.width < 0) && (bounds.height < 0) )
            hmode = vmode = dmode;
        
        checkFlipHoriz( bounds, hmode );
        checkFlipVert( bounds, vmode );
    }

    
    /**
     * Handles drags.  This moves the selected component by an amount relative
     * to where the mouse was when the drag started, or the last point that we
     * got an event for.
     */
    public void mouseDragged( MouseEvent e )
    {
        if( _selection == null )
            return;

        int cursor = cursorFromPosition( e, true );

        if( _dragStart == null )
        {
            // mousePressed failed to set start point, so set it now
            startDrag( e.getPoint(), cursor );
            return;
        }
        
        // Flag that we should constain proportions
        boolean constrain = ((e.getModifiers() & MouseEvent.CTRL_MASK) != 0);

        // Work out how much the mouse has moved
        Point dragPos = e.getPoint();
        int dx = dragPos.x - _dragStart.x;
        int dy = dragPos.y - _dragStart.y;
        
        // Work out the constrained X increment (Y is derived)
        double r = dragPos.distance( _dragStart );
        int cdx = (int) (_dragRatio < 1.0 ? r * _dragRatio : r);
        if( dx < 0 ) cdx = -cdx;
        int cdy = (int) ((_bounds.width + cdx) / _dragRatio) - _bounds.height;

        // Temp vars
        boolean moved = true;
        Rectangle bounds = _selection.getBounds();
        
        switch( _dragMode )
        {
        case Cursor.MOVE_CURSOR:
            // Move the component
            bounds.x += dx; bounds.y += dy;
            break;
        case Cursor.N_RESIZE_CURSOR:
            // Resize and move the component vertically
            bounds.y += dy; bounds.height -= dy;
            checkFlipVert( bounds, Cursor.S_RESIZE_CURSOR );
            break;
        case Cursor.NE_RESIZE_CURSOR:
            // Resize the component diagonally
            if( constrain ) { bounds.width += cdx; bounds.y += cdy; bounds.height -= cdy; }
            else { bounds.width += dx; bounds.y += dy; bounds.height -= dy; }
            checkFlipDiag( bounds, Cursor.NW_RESIZE_CURSOR, Cursor.SE_RESIZE_CURSOR, Cursor.SW_RESIZE_CURSOR );
            break;
        case Cursor.E_RESIZE_CURSOR:
            // Resize the component horizontally
            bounds.width += dx;
            checkFlipHoriz( bounds, Cursor.W_RESIZE_CURSOR );
            break;
        case Cursor.SE_RESIZE_CURSOR:
            // Resize the component diagonally
            if( constrain ) { bounds.width += cdx; bounds.height += cdy; }
            else { bounds.width += dx; bounds.height += dy; }
            checkFlipDiag( bounds, Cursor.SW_RESIZE_CURSOR, Cursor.NE_RESIZE_CURSOR, Cursor.NW_RESIZE_CURSOR );
            break;
        case Cursor.S_RESIZE_CURSOR:
            // Resize the component vertically
            bounds.height += dy;
            checkFlipVert( bounds, Cursor.N_RESIZE_CURSOR );
            break;
        case Cursor.SW_RESIZE_CURSOR:
            // Resize the component diagonally
            if( constrain ) { bounds.x += cdx; bounds.width -= cdx; bounds.height += cdy; }
            else { bounds.x += dx; bounds.width -= dx; bounds.height += dy; }
            checkFlipDiag( bounds, Cursor.SE_RESIZE_CURSOR, Cursor.NW_RESIZE_CURSOR, Cursor.NE_RESIZE_CURSOR );
            break;
        case Cursor.W_RESIZE_CURSOR:
            // Resize and move the component horizontally
            bounds.x += dx; bounds.width -= dx;
            checkFlipHoriz( bounds, Cursor.E_RESIZE_CURSOR );
            break;
        case Cursor.NW_RESIZE_CURSOR:
            // Resize the component diagonally
            if( constrain ) { bounds.x += cdx; bounds.width -= cdx; bounds.y += cdy; bounds.height -= cdy; }
            else { bounds.x += dx; bounds.width -= dx; bounds.y += dy; bounds.height -= dy; }
            checkFlipDiag( bounds, Cursor.NE_RESIZE_CURSOR, Cursor.SW_RESIZE_CURSOR, Cursor.SE_RESIZE_CURSOR );
            break;
        default:
            moved = false;
            break;
        }

        if( moved )
        {
            // Move/resize the component, which will in turn update the drag handles via an event
            _selection.setBounds( bounds );
            repaint();
        }
        
        // save the new drag position for next time
        _dragStart = dragPos;
    }
    
    
    /**
     * Handles changing the mouse cursor when placed over the selection handles
     * with no buttons pressed.
     */
    public void mouseMoved( MouseEvent e )
    {
        int newcursor = cursorFromPosition( e, false );
        
        if( _cursor != newcursor )
        {
            setCursor( Cursor.getPredefinedCursor( newcursor ) );
            _cursor = newcursor;
        }
    }   
    
    
    /**
     * Handles auto-scrolling during drags.  When the mouse is in the margin
     * around the edge of the viewport, the viewport is made to scroll in that
     * direction.
     */
    public void autoscroll( Point pt )
    {
        Insets asin = this.getAutoscrollInsets();   // The margins for comparisons
        JViewport vp = (JViewport) getParent();     // The viewport we are in
        Rectangle vrect = vp.getViewRect();         // The bounds of the viewing area
        Dimension vsz = vp.getViewSize();           // The size of the viewable area

        Point pos = new Point( vrect.x, vrect.y );  // The point we wish to move to

        // Calculate new X position
        if( pt.x < asin.left )
            pos.x = vrect.x > margin ? vrect.x - margin : 0;
        else if( pt.x > (getWidth() - asin.left - asin.right) )
            pos.x = (vsz.width - (vrect.x + vrect.width)) > margin ? vrect.x + margin : (vsz.width - vrect.width);
        
        // Calculate new Y position
        if( pt.y < asin.top )
            pos.y = vrect.y > margin ? vrect.y - margin : 0;
        else if( pt.y > (getHeight() - asin.top - asin.bottom) )
            pos.y = (vsz.height - (vrect.y + vrect.height)) > margin ? vrect.y + margin : (vsz.height - vrect.height);

        // Only move if position changed
        if( (vrect.x != pos.x) || (vrect.y != pos.y) )
            vp.setViewPosition( pos );
    }
    
    
    /**
     * Returns the margins of the viewport, relative to the bounds of the panel.
     */
    public Insets getAutoscrollInsets()
    {
        Rectangle outer = this.getBounds();
        Rectangle inner = getParent().getBounds();
        return new Insets( inner.y - outer.y + margin, inner.x - outer.x + margin,
                           outer.height - inner.height - inner.y + outer.y + margin,
                           outer.width - inner.width - inner.x + outer.x + margin );
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Inner classes

    /**
     * Simple inner class to look after updating selection rectangles in
     * repsonse to ComponentEvents from any of the (many) components we are
     * listening to.
     */
    private class RectangleUpdater
        extends ComponentAdapter
    {
        public void componentMoved( ComponentEvent e ) {
            updateSelectionRectangles();
        }
        public void componentResized( ComponentEvent e ) {
            updateSelectionRectangles();
        }
    }
    
    RectangleUpdater rectangleUpdater = new RectangleUpdater();
    
    
    /**
     * Simple inne class to look after setting the selection to the component
     * which has gained the input focus.
     */
    private class SelectionUpdater
        extends FocusAdapter
    {
        public void focusGained( FocusEvent e ) {
            if( !e.isTemporary() ) {
                System.out.println( "Focus Gained: " + e.getComponent() );
                //setSelection( e.getComponent() );
            }
        }
    }
    
    SelectionUpdater selectionUpdater = new SelectionUpdater();
    
    
    /**
     * More complex inner class which deals with components being added and
     * removed from any of the (many) containers we are listening to.  These
     * containers may already contain components/containers, so these are
     * recursively dealt with.
     */
    private class ComponentUpdater
        implements ContainerListener
    {
        public void componentAdded( ContainerEvent e )
        {
            Component c = e.getChild();
            c.addComponentListener( rectangleUpdater );
            c.addFocusListener( selectionUpdater );
            if( c instanceof Container ) {
                ((Container) c).addContainerListener( this );
                recursiveAddListeners( (Container) c );
            }
        }
        
        public void componentRemoved( ContainerEvent e )
        {
            Component c = e.getChild();
            c.removeComponentListener( rectangleUpdater );
            c.removeFocusListener( selectionUpdater );
            if( c instanceof Container ) {
                ((Container) c).removeContainerListener( this );
            }
        }
        
        private void recursiveAddListeners( Container cont )
        {
            for( int n = cont.getComponentCount() - 1; n >= 0; n-- ) {
                Component c = cont.getComponent( n );
                c.addComponentListener( rectangleUpdater );
                c.addFocusListener( selectionUpdater );
                if( c instanceof Container ) {
                    ((Container) c).addContainerListener( this );
                    recursiveAddListeners( (Container) c );
                }
            }
        }
        
        private void recursiveRemoveListeners( Container cont )
        {
            for( int n = cont.getComponentCount() - 1; n >= 0; n-- ) {
                Component c = cont.getComponent( n );
                c.removeComponentListener( rectangleUpdater );
                c.removeFocusListener( selectionUpdater );
                if( c instanceof Container ) {
                    ((Container) c).removeContainerListener( this );
                    recursiveRemoveListeners( (Container) c );
                }
            }
        }
    }
    
    
    /**
     * DropTarget class for handling incoming data being droppped onto the
     * editing area.  Looks after highlighting the component to get the drop
     * and the action which occurs when the drop happens.
     */
    private class DPDropTarget
        implements DropTargetListener
    {
        private DropTarget target;
        private DateablePanel panel;
        private DataFlavor classDataFlavor;
        private Point oldpos;
        private boolean accept;
        
        public DPDropTarget( DateablePanel dp )
        {
            panel = dp;
            target = new DropTarget( dp, this );

            try {
                classDataFlavor = new DataFlavor( DataFlavor.javaJVMLocalObjectMimeType + "; class=java.lang.Class" );
            } catch( ClassNotFoundException e ) {
                e.printStackTrace();
            }
        }
        
        private void acceptDrag( DropTargetDragEvent dtde, int action )
        {
            accept = true;
            dtde.acceptDrag( DnDConstants.ACTION_COPY );
        }

        private void rejectDrag( DropTargetDragEvent dtde )
        {
            accept = false;
            dtde.rejectDrag();
        }
        
        private Component dropSiteFromLocationForFileList( Point loc )
        {
            // Get innermost component
            Component leaf = findComponentAt( (int) (_iscale * loc.x), (int) (_iscale * loc.y) );
            
            // Is it anything we want?
            if( (leaf instanceof MonthImage)
                || (leaf instanceof CalendarPanel) )
                return leaf;
            
            // No, so try its parent
            Component parent = leaf.getParent();
            
            // Is parent acceptable?
            if( (parent != null) && (parent instanceof CalendarPanel) )
                return parent;
            
            return null;
        }

        
        public void dragEnter( DropTargetDragEvent dtde )
        {
            if( dtde.isDataFlavorSupported( DataFlavor.javaFileListFlavor ) )
            {
                // File list, so accept it
                acceptDrag( dtde, DnDConstants.ACTION_COPY );

                // Turn on the rectangle around the component we are going to give the drop to
                _dropSite = dropSiteFromLocationForFileList( dtde.getLocation() );
                repaint();
                return;
            }
            else if( dtde.isDataFlavorSupported( classDataFlavor ) )
            {
                // Class object, so accept it if we have a valid parent
                if( parentFromLocationForBeanInsert( dtde.getLocation() ) != null )
                {
                    acceptDrag( dtde, DnDConstants.ACTION_COPY );
                    return;
                }
            }

            // We got to here, so all we can do is reject it
            rejectDrag( dtde );
        }
        
        public void dragExit( DropTargetEvent dte )
        {
            // Clear the rectangle around the component we were going to give the drop to
            _dropSite = null;
            repaint();
        }
        
        public void dragOver( DropTargetDragEvent dtde )
        {
            // De-bounce events - we are only interested if the cursor has actually moved!
            if( (oldpos != null) && oldpos.equals( dtde.getLocation() ) )
            {
                if( accept ) dtde.acceptDrag( DnDConstants.ACTION_COPY );
                else dtde.rejectDrag();
                return;
            }

            oldpos = dtde.getLocation();
            
            if( dtde.isDataFlavorSupported( DataFlavor.javaFileListFlavor ) )
            {
                // File list, so accept it
                acceptDrag( dtde, DnDConstants.ACTION_COPY );

                Component newds = dropSiteFromLocationForFileList( dtde.getLocation() );

                if( _dropSite != newds )
                {
                    _dropSite = newds;
                    repaint();
                }
                
                return;
            }
            else if( dtde.isDataFlavorSupported( classDataFlavor ) )
            {
                // Class object, so accept it if we have a valid parent
                if( parentFromLocationForBeanInsert( dtde.getLocation() ) != null )
                {
                    acceptDrag( dtde, DnDConstants.ACTION_COPY );
                    return;
                }
            }

            // We got to here, so all we can do is reject it
            rejectDrag( dtde );
        }
        
        public void drop( DropTargetDropEvent dtde )
        {
            // Clear the rectangle around the component we gave the drop to
            _dropSite = null;
            repaint();
            
            try
            {
                Transferable t = dtde.getTransferable();
                
                // DEBUG
                DataFlavor[] flavors = t.getTransferDataFlavors();
                for( int n = 0; n < flavors.length; n ++ )
                    System.out.println( "flavor[" + n + "] = " + flavors[n].getMimeType() );
                // END DEBUG
                
                // Handle any files dragged from a file browser (e.g. Windows Explorer)
                if( t.isDataFlavorSupported( DataFlavor.javaFileListFlavor ) )
                {
                    dropFileList( dtde );
                    return;
                }
                else if( t.isDataFlavorSupported( classDataFlavor ) )
                {
                    dropClassObject( dtde );
                    return;
                }
            }
            catch( Exception e )
            {
                e.printStackTrace();
            }
            
            // If we didn't handle it, reject it!
            dtde.rejectDrop();
        }
        
        public void dropActionChanged( DropTargetDragEvent dtde )  {}
        
        
        private void dropFileList( DropTargetDropEvent dtde )
        {
            dtde.acceptDrop( DnDConstants.ACTION_COPY );
            String msg = "Unknown_Error";

            try
            {
                Transferable t = dtde.getTransferable();
                Component actualDropSite = dropSiteFromLocationForFileList( dtde.getLocation() );
                List data = (List) t.getTransferData( DataFlavor.javaFileListFlavor );
                
                // More than one file not handled yet!
                if( data.size() == 1 )
                {
                    File file = (File) data.get( 0 );
                    String name = file.getName();

                    if( name.endsWith( ".cal" ) || name.endsWith( ".xml" ) )
                    {
                        // A template file - load it provided we land on anything but a MonthImage
                        if( (actualDropSite == null) || (actualDropSite instanceof CalendarPanel) )
                        {
                            // Load event
                            fireDateableSelectionEvent( new DateableSelectionEvent( DateablePanel.this, file ) );
                            dtde.dropComplete( true );
                            return;
                        }
                        else
                            msg = "Import_Template_Error";
                    }
                    else if( name.endsWith( ".jpg" ) || name.endsWith( ".jpeg" )
                            || name.endsWith( ".gif" ) || name.endsWith( ".png" ) )
                    {
                        msg = "Import_Image_Error";
                        
                        // An image file - figure out what we landed on:
                        if( actualDropSite != null )
                        {
                            if( actualDropSite instanceof MonthImage )
                            {
                                // Set current image if a MonthImage
                                ((MonthImage) actualDropSite).setImages( _month, file.toString() );
                                dtde.dropComplete( true );
                                return;
                            }
                            else if( actualDropSite instanceof CalendarPanel )
                            {
                                // Create new static image
                                // TODO: Create StaticImage
                                dtde.dropComplete( true );
                                return;
                            }
                        }
                    }
                    else
                        msg = "Import_FileType_Error";
                }
                else
                    msg = "Import_Multiple_Error";
            }
            catch( Exception e )
            {
                msg = "Import_Exception_Error";
                e.printStackTrace();
            }

            // We accepted it, but couldn't do anything with it!
            dtde.dropComplete( false );
            
            // Report the error message
            String title = calendarBeans.getString( "Import_Error_Title" );
            String message = calendarBeans.getString( msg );
            JOptionPane.showMessageDialog( DateablePanel.this, message, title, JOptionPane.ERROR_MESSAGE );
        }
        
        
        private void dropClassObject( DropTargetDropEvent dtde )
        {
            Point loc = dtde.getLocation();
            Container parent = parentFromLocationForBeanInsert( loc );

            if( parent == null )
            {
                dtde.rejectDrop();
                return;
            }

            dtde.acceptDrop( DnDConstants.ACTION_COPY );
            String msg = "Unknown_Error";

            try
            {
                Transferable t = dtde.getTransferable();
                Class beanClass = (Class) t.getTransferData( classDataFlavor );

                insertBean( beanClass, parent, loc.x, loc.y );
                dtde.dropComplete( true );
                return;
            }
            catch( Exception e )
            {
                msg = "Import_Exception_Error";
                e.printStackTrace();
            }

            // We accepted it, but couldn't do anything with it!
            dtde.dropComplete( false );

            // Report the error message
            String title = calendarBeans.getString( "Import_Error_Title" );
            String message = calendarBeans.getString( msg );
            JOptionPane.showMessageDialog( DateablePanel.this, message, title, JOptionPane.ERROR_MESSAGE );
        }
    }
}